#include <Ultrasonic.h>
#include <AFMotor.h>

// Create motor objects
AF_DCMotor motor1(1);  // Motor 1
AF_DCMotor motor2(2);  // Motor 2
AF_DCMotor motor3(3);  // Motor 3
AF_DCMotor motor4(4);  // Motor 4

// Define pins for ultrasonic sensors
const int trigPin1 = 26;  // Trigger pin for front sensor
const int echoPin1 = 27;  // Echo pin for front sensor
const int trigPin2 = 28;  // Trigger pin for left sensor
const int echoPin2 = 29;  // Echo pin for left sensor
const int trigPin3 = 30;  // Trigger pin for back sensor
const int echoPin3 = 31;  // Echo pin for back sensor
const int trigPin4 = 32;  // Trigger pin for right sensor
const int echoPin4 = 33;  // Echo pin for right sensor

// Create global instances of Ultrasonic for each sensor
Ultrasonic frontSensor(trigPin1, echoPin1);
Ultrasonic leftSensor(trigPin2, echoPin2);
Ultrasonic backSensor(trigPin3, echoPin3);
Ultrasonic rightSensor(trigPin4, echoPin4);

int dis = 20;          // Distance threshold for obstacle detection
int mS1 = 200;  // Motor speed
int mS2 = 200;  
int mS3 = 200;  
int mS4 = 200;  
int index = 0;

#define ARRAY_SIZE 4

// Initialize the circular array
int circularArray[ARRAY_SIZE] = {0, 1, 2, 3};

// PID control variables
float Kp = 5.0, Ki = 0.005, Kd = 0.005;
 float previous_error , integral;

void setup() {
  // Initialize Serial communication
  Serial.begin(9600);
}

// Function to clamp speed values
int clampSpeed(int speed) {
  if (speed < 0) {
    return 145;
  } else if (speed > 255) {
    return 255;
  }
  return speed;
}

void turnLeft() {
  Serial.println("Turning left...");
  index = (index + 1) % 4;

  // Update sensor assignments
  Ultrasonic tempSensor = frontSensor;
  frontSensor = leftSensor;
  leftSensor = backSensor;
  backSensor = rightSensor;
  rightSensor = tempSensor;
}

void turnRight() {
  Serial.println("Turning right...");
  index = (index + 3) % 4;

  // Update sensor assignments
  Ultrasonic tempSensor = frontSensor;
  frontSensor = rightSensor;
  rightSensor = backSensor;
  backSensor = leftSensor;
  leftSensor = tempSensor;
}

void reverse() {
  Serial.println("Reversing...");
  index = (index + 2) % 4;

  // Update sensor assignments
  Ultrasonic tempSensor = frontSensor;
  frontSensor = backSensor;
  backSensor = tempSensor;
  tempSensor = leftSensor;
  leftSensor = rightSensor;
  rightSensor = tempSensor;
}

void nm(int dif) {
  // Move north
  motor1.setSpeed(clampSpeed(mS1-dif));
  motor1.run(FORWARD);

  motor2.setSpeed(clampSpeed(mS2+dif));
  motor2.run(FORWARD);

  motor3.setSpeed(clampSpeed(mS3+dif));
  motor3.run(FORWARD);

  motor4.setSpeed(clampSpeed(mS4-dif));
  motor4.run(FORWARD);
}

void wm(int dif) {
  // Move south
  motor1.setSpeed(clampSpeed(mS1-dif));
  motor1.run(FORWARD);

  motor2.setSpeed(clampSpeed(mS2-dif));
  motor2.run(BACKWARD);

  motor3.setSpeed(clampSpeed(mS3+dif));
  motor3.run(FORWARD);

  motor4.setSpeed(clampSpeed(mS4+dif));
  motor4.run(BACKWARD);
}

void sm(int dif) {
  // Move west
  motor1.setSpeed(clampSpeed(mS1+dif));
  motor1.run(BACKWARD);

  motor2.setSpeed(clampSpeed(mS2-dif));
  motor2.run(BACKWARD);

  motor3.setSpeed(clampSpeed(mS3-dif));
  motor3.run(BACKWARD);

  motor4.setSpeed(clampSpeed(mS4+dif));
  motor4.run(BACKWARD);
}

void em(int dif) {
  // Move east
  motor1.setSpeed(clampSpeed(mS1+dif));
  motor1.run(BACKWARD);

  motor2.setSpeed(clampSpeed(mS2+dif));
  motor2.run(FORWARD);

  motor3.setSpeed(clampSpeed(mS3-dif));
  motor3.run(BACKWARD);

  motor4.setSpeed(clampSpeed(mS4-dif));
  motor4.run(FORWARD);
}

void stopMotor() {
  // Stop all motors
  motor1.run(RELEASE);
  motor2.run(RELEASE);
  motor3.run(RELEASE);
  motor4.run(RELEASE);
}

float pid(int leftDistance, int rightDistance) {
  // PID control to maintain parallel to walls
  float error = leftDistance - rightDistance;
  if (-20 < error < 20) {
     integral += error;
    float derivative = error - previous_error;
    float adjustment = Kp * error + Ki * integral + Kd * derivative;
     previous_error = error;
    return adjustment * 5;
  } else {
    return 0;
  }
}

void loop() {
  // Measure distances
  int frontDistance = frontSensor.read();
  int leftDistance = leftSensor.read();
  int backDistance = backSensor.read();
  int rightDistance = rightSensor.read();

  // Print distances to Serial Monitor
  Serial.println("Front Sensor: " + String(frontDistance) + " cm");
  Serial.println("Left Sensor: " + String(leftDistance) + " cm");
  Serial.println("Back Sensor: " + String(backDistance) + " cm");
  Serial.println("Right Sensor: " + String(rightDistance) + " cm");

  if (frontDistance < dis && leftDistance > dis) {
    stopMotor();
    delay(500);
    turnLeft();
    previous_error = 0, integral = 0;
  } else if (frontDistance < dis && leftDistance < dis && rightDistance > dis) {
    stopMotor();
    delay(500);
    turnRight();
    previous_error = 0, integral = 0;
  } else if (frontDistance < dis && leftDistance < dis && rightDistance < dis) {
    stopMotor();
    delay(500);
    reverse();
    previous_error = 0, integral = 0;
  } else {
   

    int dif = pid(leftDistance, rightDistance); // Call PID and use returned value
    switch (index) {
      case 0:
        nm(dif);
        break;
      case 1:
        em(dif);
        break;
      case 2:
        sm(dif);
        break;
      case 3:
        wm(dif);
        break;
    }
 
    
  }
}
